from collections import defaultdict, OrderedDict
from copy import deepcopy
from datetime import datetime
from functools import lru_cache

from tabulate import tabulate

from vigorish.cli.components.models import DisplayTable
from vigorish.cli.components.table_viewer import TableViewer
from vigorish.config.database import PitchAppScrapeStatus, PlayerId, Team
from vigorish.data.all_player_data import AllPlayerData
from vigorish.enums import DataSet, DefensePosition, VigFile
from vigorish.util.dt_format_strings import DT_AWARE, DT_AWARE_VERBOSE
from vigorish.util.exceptions import ScrapedDataException
from vigorish.util.list_helpers import flatten_list2d, group_and_sort_dict_list, make_chunked_list
from vigorish.util.regex import INNING_LABEL_REGEX
from vigorish.util.result import Result
from vigorish.util.string_helpers import (
    get_bbref_team_id,
    inning_number_to_string,
    validate_at_bat_id,
    validate_bbref_game_id,
    wrap_text,
)

AT_BAT_TABLE_MAX_ROWS = 8
AT_BAT_DESC_MAX_WIDTH = 35


class AllGameData:
    def __init__(self, app, bbref_game_id):
        # TODO: Create method to get pitch type counts for pitch app
        self.app = app
        self.db_engine = app["db_engine"]
        self.db_session = app["db_session"]
        self.scraped_data = app["scraped_data"]
        combined_data = self.scraped_data.get_combined_game_data(bbref_game_id)
        if not combined_data:
            raise ScrapedDataException(
                file_type=VigFile.COMBINED_GAME_DATA, data_set=DataSet.ALL, url_id=bbref_game_id
            )
        self.bbref_game_id = combined_data["bbref_game_id"]
        self.pitchfx_vs_bbref_audit = combined_data["pitchfx_vs_bbref_audit"]
        self.game_meta_info = combined_data["game_meta_info"]
        self.away_team_data = combined_data["away_team_data"]
        self.home_team_data = combined_data["home_team_data"]
        self.innings_list = combined_data["play_by_play_data"]
        self.duplicate_guids = combined_data["duplicate_guids"]
        self.removed_pitchfx = combined_data["removed_pitchfx"]
        self.invalid_pitchfx = combined_data["invalid_pitchfx"]
        self.player_id_dict = combined_data["player_id_dict"]
        self.all_player_mlb_ids = [id_map["mlb_id"] for id_map in self.player_id_dict.values()]
        self.game_date = validate_bbref_game_id(self.bbref_game_id).value["game_date"]

    @property
    def away_team_id(self):
        return self.away_team_data["team_id_br"]

    @property
    def home_team_id(self):
        return self.home_team_data["team_id_br"]

    @property
    def away_team(self):
        return Team.find_by_team_id_and_year(
            self.db_session, self.away_team_id, self.game_date.year
        )

    @property
    def home_team(self):
        return Team.find_by_team_id_and_year(
            self.db_session, self.home_team_id, self.game_date.year
        )

    @property
    def game_datetime(self):
        return datetime.strptime(self.game_meta_info["game_date_time_str"], DT_AWARE)

    @property
    def bat_stats_player_ids(self):
        batter_mlb_ids = [
            int(validate_at_bat_id(at_bat_id).value["batter_mlb_id"])
            for at_bat_id in self.get_at_bat_map().keys()
        ]
        return list(set(batter_mlb_ids))

    @property
    def pitch_stats_player_ids(self):
        return [mlb_id for mlb_id in self.get_pitch_stat_map().keys()]

    @property
    def team_data(self):
        return {
            self.away_team_id: self.away_team_data,
            self.home_team_id: self.home_team_data,
        }

    @property
    def bat_boxscore(self):
        return {
            self.away_team_id: self.create_bat_boxscore(self.away_team_id),
            self.home_team_id: self.create_bat_boxscore(self.home_team_id),
        }

    @lru_cache(maxsize=None)
    def create_bat_boxscore(self, team_id):
        team_data = self.team_data[team_id]
        batter_box = OrderedDict()
        for slot in team_data["starting_lineup"]:
            mlb_id = self.get_player_id_map(bbref_id=slot["player_id_br"]).mlb_id
            batter_box[slot["bat_order"]] = self.get_bat_boxscore_for_player(
                mlb_id, slot["def_position"], team_data
            )
        lineup_player_ids = [bat_boxscore["mlb_id"] for bat_boxscore in batter_box.values()]
        sub_player_ids = [
            mlb_id
            for mlb_id in self.get_all_player_ids_by_team(team_data["team_id_br"])
            if mlb_id in self.bat_stats_player_ids and mlb_id not in lineup_player_ids
        ]
        for num, sub_id in enumerate(sub_player_ids, start=1):
            batter_box[f"BN{num}"] = self.get_bat_boxscore_for_player(sub_id, "BN", team_data)
        return batter_box

    def get_player_id_map(self, mlb_id=None, bbref_id=None):
        if not mlb_id and not bbref_id:
            return None
        return (
            PlayerId.find_by_mlb_id(self.db_session, mlb_id)
            if mlb_id
            else PlayerId.find_by_bbref_id(self.db_session, bbref_id)
        )

    def get_all_player_ids_by_team(self, team_id):
        return [
            player_dict["mlb_id"]
            for player_dict in self.player_id_dict.values()
            if player_dict["team_id_bbref"] == team_id
        ]

    def get_bat_boxscore_for_player(self, mlb_id, def_position, team_data):
        player_id = self.get_player_id_map(mlb_id=mlb_id)
        bat_stats = self.get_bat_stats(mlb_id).value
        (at_bats, details) = parse_bat_stats_for_game(bat_stats)
        return {
            "team_id": team_data["team_id_br"],
            "name": player_id.mlb_name,
            "mlb_id": player_id.mlb_id,
            "bbref_id": player_id.bbref_id,
            "def_position": DefensePosition.from_abbrev(def_position),
            "at_bats": at_bats,
            "bat_stats": details,
            "stats_to_date": parse_bat_stats_to_date(bat_stats),
            "at_bat_viewer": self.view_valid_at_bats_for_batter(mlb_id).value,
        }

    @property
    def pitch_boxscore(self):
        return {
            self.away_team_id: self.create_pitch_boxscore(self.away_team_id),
            self.home_team_id: self.create_pitch_boxscore(self.home_team_id),
        }

    @lru_cache(maxsize=None)
    def create_pitch_boxscore(self, team_id):
        team_data = self.team_data[team_id]
        pitcher_box = OrderedDict()
        pitcher_ids = [
            mlb_id
            for mlb_id in self.get_all_player_ids_by_team(team_data["team_id_br"])
            if mlb_id in self.pitch_stats_player_ids
        ]
        rp_count = 0
        for mlb_id in self.get_pitcher_app_order(pitcher_ids):
            player_box = self.get_pitch_boxscore_for_player(mlb_id, team_data)
            if player_box["pitch_app_type"] == "SP":
                dict_index = "SP"
            else:
                rp_count += 1
                dict_index = f"RP{rp_count}"
            pitcher_box[dict_index] = player_box
        return pitcher_box

    def get_pitcher_app_order(self, pitcher_ids):
        pitcher_app_dict = {}
        for mlb_id in pitcher_ids:
            at_bats = self.get_valid_at_bats_for_pitcher(mlb_id).value
            first_pitch = min(at_bat["pbp_table_row_number"] for at_bat in at_bats)
            pitcher_app_dict[mlb_id] = first_pitch
        return sorted(pitcher_ids, key=lambda x: pitcher_app_dict[x])

    def get_pitch_boxscore_for_player(self, mlb_id, team_data):
        player_id = self.get_player_id_map(mlb_id=mlb_id)
        pitch_stats = self.get_pitch_app_stats(mlb_id).value
        game_stats = parse_pitch_app_stats(pitch_stats)
        is_starter = "GS" in game_stats
        return {
            "team_id": team_data["team_id_br"],
            "name": player_id.mlb_name,
            "mlb_id": mlb_id,
            "bbref_id": player_id.bbref_id,
            "pitch_app_type": "SP" if is_starter else "RP",
            "game_results": game_stats,
            "at_bat_viewer": self.view_valid_at_bats_for_pitcher(mlb_id).value,
        }

    @lru_cache(maxsize=None)
    def get_valid_at_bats(self):
        return [
            at_bat_data
            for inning_data in self.innings_list
            for at_bat_data in inning_data["inning_events"]
        ]

    @lru_cache(maxsize=None)
    def get_invalid_at_bats(self):
        return [
            at_bat_data
            for inning_dict in self.invalid_pitchfx.values()
            for at_bat_data in inning_dict.values()
        ]

    def get_all_at_bats(self):
        return self.get_valid_at_bats() + self.get_invalid_at_bats()

    def get_at_bat_map(self):
        return {at_bat["at_bat_id"]: at_bat for at_bat in self.get_all_at_bats()}

    @lru_cache(maxsize=None)
    def get_all_pitch_stats(self):
        return self.away_team_data["pitching_stats"] + self.home_team_data["pitching_stats"]

    def get_pitch_stat_map(self):
        return {
            pitch_stats["pitcher_id_mlb"]: pitch_stats for pitch_stats in self.get_all_pitch_stats()
        }

    @lru_cache(maxsize=None)
    def get_all_bat_stats(self):
        bat_stats = self.away_team_data["batting_stats"] + self.home_team_data["batting_stats"]
        return [bat_stat for bat_stat in bat_stats if bat_stat["total_plate_appearances"]]

    def get_bat_stat_map(self):
        return {bat_stats["batter_id_mlb"]: bat_stats for bat_stats in self.get_all_bat_stats()}

    def view_valid_at_bats_for_batter(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        at_bats = self.get_valid_at_bats_for_batter(mlb_id).value
        batter_tables = []
        for num, at_bat in enumerate(at_bats, start=1):
            player_name = self.get_player_id_map(mlb_id=mlb_id).mlb_name
            heading = f"At Bat #{num}/{len(at_bats)} for {player_name} in Game {self.bbref_game_id}"
            table_list = self.create_at_bat_table_list(at_bat["at_bat_id"], heading)
            batter_tables.extend(table_list)
        table_viewer = self.create_table_viewer(batter_tables)
        return Result.Ok(table_viewer)

    def validate_mlb_id(self, mlb_id):
        if not isinstance(mlb_id, int):
            try:
                mlb_id = int(mlb_id)
            except ValueError:
                return Result.Fail(f'"{mlb_id}" could not be parsed as a valid int')
        if mlb_id not in self.all_player_mlb_ids:
            error = f"Error: Player MLB ID: {mlb_id} did not appear in game {self.bbref_game_id}"
            return Result.Fail(error)
        return Result.Ok(mlb_id)

    def get_valid_at_bats_for_batter(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        at_bats = [ab for ab in self.get_valid_at_bats() if ab["batter_id_mlb"] == mlb_id]
        return Result.Ok(at_bats)

    @lru_cache(maxsize=None)
    def create_at_bat_table_list(self, at_bat_id, heading=None):
        at_bat = self.get_at_bat_map().get(at_bat_id)
        message = self.get_at_bat_details(at_bat)
        pitch_seq_desc = format_pitch_sequence_description(at_bat["pitch_sequence_description"])
        chunked_list = make_chunked_list(pitch_seq_desc, chunk_size=AT_BAT_TABLE_MAX_ROWS)
        return [
            DisplayTable(tabulate(chunk, tablefmt="fancy_grid"), heading, message)
            for chunk in chunked_list
        ]

    def get_at_bat_details(self, at_bat):
        pitch_app_stats = self.get_pitch_app_stats(at_bat["pitcher_id_mlb"]).value
        bat_stats = self.get_bat_stats(at_bat["batter_id_mlb"]).value
        return get_at_bat_details(at_bat, pitch_app_stats, bat_stats)

    def get_pitch_app_stats(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        pitch_stats = self.get_pitch_stat_map().get(mlb_id)
        return Result.Ok(pitch_stats)

    def get_bat_stats(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        bat_stats = self.get_bat_stat_map().get(mlb_id)
        return Result.Ok(bat_stats["bbref_data"]) if bat_stats else Result.Ok({})

    def create_table_viewer(self, table_list, table_color="bright_cyan"):
        return TableViewer(
            table_list=table_list,
            prompt="Press Enter to return to previous menu",
            confirm_only=True,
            table_color=table_color,
            heading_color="bright_yellow",
            message_color=None,
        )

    def view_valid_at_bats_for_pitcher(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        at_bats = self.get_valid_at_bats_for_pitcher(mlb_id).value
        at_bats_by_inning = group_and_sort_dict_list(at_bats, "inning_id", "pbp_table_row_number")
        at_bat_tables_by_inning = {}
        all_at_bat_tables = []
        for inning_id, at_bats in at_bats_by_inning.items():
            inning = inning_id[-5:]
            at_bat_tables = []
            for num, at_bat in enumerate(at_bats, start=1):
                player_name = self.get_player_id_map(mlb_id=mlb_id).mlb_name
                heading = f"Inning: {inning}, At Bat #{num}/{len(at_bats)}, P: {player_name}"
                table_list = self.create_at_bat_table_list(at_bat["at_bat_id"], heading)
                at_bat_tables.extend(table_list)
            at_bat_tables_by_inning[inning] = at_bat_tables
            all_at_bat_tables.extend(at_bat_tables)

        innings_viewer = {"ALL": self.create_table_viewer(all_at_bat_tables)}
        for inning, at_bat_tables in at_bat_tables_by_inning.items():
            innings_viewer[inning] = self.create_table_viewer(at_bat_tables)
        return Result.Ok(innings_viewer)

    def get_valid_at_bats_for_pitcher(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        at_bats = [
            at_bat for at_bat in self.get_valid_at_bats() if at_bat["pitcher_id_mlb"] == mlb_id
        ]
        return Result.Ok(at_bats)

    def view_at_bats_by_inning(self):
        all_at_bat_tables = []
        at_bat_tables_by_inning = {}
        for inning_dict in self.get_innings_sorted():
            inning = inning_dict["inning_id"][-5:]
            inning_tables = []
            for num, at_bat in enumerate(inning_dict["at_bats"], start=1):
                heading = f"Inning: {inning}, At Bat #{num}/{len(inning_dict['at_bats'])}"
                table_list = self.create_at_bat_table_list(at_bat["at_bat_id"], heading)
                inning_tables.extend(table_list)
            at_bat_tables_by_inning[inning] = inning_tables
            all_at_bat_tables.extend(inning_tables)

        innings_viewer = {"ALL": self.create_table_viewer(all_at_bat_tables)}
        for inning, at_bat_tables in at_bat_tables_by_inning.items():
            innings_viewer[inning] = self.create_table_viewer(at_bat_tables)
        return Result.Ok(innings_viewer)

    def get_innings_sorted(self):
        at_bats_by_inning = group_and_sort_dict_list(
            self.get_valid_at_bats(), "inning_id", "pbp_table_row_number"
        )
        innings_unsorted = [
            {"inning_id": inning_id, "at_bats": at_bats}
            for inning_id, at_bats in at_bats_by_inning.items()
        ]
        return get_innings_sorted(innings_unsorted)

    def view_game_meta_info(self):
        meta_info = deepcopy(self.game_meta_info)
        meta_info.pop("umpires")
        meta_info.pop("game_time_hour")
        meta_info.pop("game_time_minute")
        game_start_time = meta_info.pop("game_date_time_str")
        table_rows = [[name, value] for name, value in meta_info.items()]
        table_rows.insert(0, ["game_start_time", game_start_time])
        table = tabulate(table_rows)
        heading = f"Meta Information for game {self.bbref_game_id}"
        return self.create_table_viewer([DisplayTable(table, heading)])

    def view_invalid_at_bats(self):
        pass

    def compare_removed_pitchfx(self):
        pass

    @lru_cache(maxsize=None)
    def get_all_pitchfx(self):
        pfx_dict = defaultdict(list)
        all_pitchfx = [at_bat["pitchfx"] for at_bat in self.get_all_at_bats()]
        all_pitchfx.extend(self.get_duplicate_guid_pfx())
        all_pitchfx.extend(self.get_removed_pfx())
        for pfx in flatten_list2d(all_pitchfx):
            pfx_dict[pfx["pitch_app_id"]].append(pfx)
        return pfx_dict

    def get_duplicate_guid_pfx(self):
        return [
            at_bat["pitchfx"]
            for inning_dict in self.duplicate_guids.values()
            for at_bat in inning_dict.values()
        ]

    def get_removed_pfx(self):
        return [
            at_bat["pitchfx"]
            for inning_dict in self.removed_pitchfx.values()
            for at_bat in inning_dict.values()
        ]

    def get_pitch_mix(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        pitch_app_id = f"{self.bbref_game_id}_{mlb_id}"
        pitcher_data = AllPlayerData(self.app, mlb_id)
        pitch_app = PitchAppScrapeStatus.find_by_pitch_app_id(self.db_session, pitch_app_id)
        return {
            "pitch_app": pitch_app.pitch_mix,
            "by_year": pitcher_data.pitch_mix_by_year(),
            "all": pitcher_data.pitch_mix(),
            "bat_r": pitcher_data.pitch_mix_right(),
            "bat_l": pitcher_data.pitch_mix_left(),
        }

    def get_matchup_details(self):
        away_record = get_team_record_for_linescore(self.away_team_data)
        home_record = get_team_record_for_linescore(self.home_team_data)
        matchup = f"{self.away_team.name} ({away_record}) vs {self.home_team.name} ({home_record})"
        game_time = self.game_datetime.strftime(DT_AWARE_VERBOSE)
        return f"{game_time}\n{matchup}\n"

    def get_linescore(self):
        return get_linescore_tables(self.away_team_data, self.home_team_data, self.innings_list)


def get_linescore_tables(away_team_data, home_team_data, innings_list):
    table_list = []
    (away_team_runs_chunked, home_team_runs_chunked) = get_runs_by_inning(innings_list)
    for i in range(len(away_team_runs_chunked)):
        start_inning = i * 9 + 1
        last_chunk = 1 == len(away_team_runs_chunked) - 1
        table = create_linescore_table(
            away_team_data["team_id_br"],
            away_team_runs_chunked[i],
            get_team_totals_for_linescore(away_team_data),
            home_team_data["team_id_br"],
            home_team_runs_chunked[i],
            get_team_totals_for_linescore(home_team_data),
            start_inning,
            last_chunk,
        )
        table_list.append(table)
    return "\n\n".join(table_list)


def get_team_record_for_linescore(team_data):
    team_wins = team_data["total_wins_before_game"]
    team_losses = team_data["total_losses_before_game"]
    return f"{team_wins}-{team_losses}"


def get_team_totals_for_linescore(team_data):
    team_total_runs = team_data["total_runs_scored_by_team"]
    team_hits = team_data["total_hits_by_team"]
    team_errors = team_data["total_errors_by_team"]
    team_totals = [team_total_runs, team_hits, team_errors]
    return team_totals


def get_runs_by_inning(innings_list):
    is_top_half = True
    away_team_runs = []
    home_team_runs = []
    for inning in innings_list:
        total_runs_this_inning = get_total_runs_in_inning(inning["inning_events"])
        if is_top_half:
            away_team_runs.append(total_runs_this_inning)
        else:
            home_team_runs.append(total_runs_this_inning)
        is_top_half = not is_top_half
    if len(away_team_runs) > len(home_team_runs):
        home_team_runs.append("X")
    away_team_runs_chunked = make_chunked_list(away_team_runs, chunk_size=9)
    home_team_runs_chunked = make_chunked_list(home_team_runs, chunk_size=9)
    return (away_team_runs_chunked, home_team_runs_chunked)


def get_total_runs_in_inning(inning_at_bats):
    return sum(ab["runs_outs_result"].count("R") for ab in inning_at_bats)


def create_linescore_table(
    away_team_id,
    away_team_runs_by_inning,
    away_team_totals,
    home_team_id,
    home_team_runs_by_inning,
    home_team_totals,
    start_inning,
    last_chunk,
):
    end_inning = start_inning + len(away_team_runs_by_inning)
    inning_numbers = get_linescore_inning_numbers(start_inning, end_inning)
    if start_inning != 1 and last_chunk:
        away_team_runs_by_inning = add_filler_columns(away_team_runs_by_inning)
        home_team_runs_by_inning = add_filler_columns(home_team_runs_by_inning)
        inning_numbers = add_filler_columns(inning_numbers)
    total_headers = [f"{' '*(4-len(letter))}{letter}" for letter in ["R", "H", "E"]]
    column_headers = [f"{' '*(4-len(str(num)))}{num}" for num in inning_numbers]
    away_numbers = [f"{' '*(4-len(str(num)))}{num}" for num in away_team_runs_by_inning]
    away_totals = [f"{' '*(4-len(str(num)))}{num}" for num in away_team_totals]
    home_numbers = [f"{' '*(4-len(str(num)))}{num}" for num in home_team_runs_by_inning]
    home_totals = [f"{' '*(4-len(str(num)))}{num}" for num in home_team_totals]
    return (
        f"   {''.join(column_headers)}{''.join(total_headers)}\n"
        f"----{'----'*9}{'----'*3}\n"
        f"{away_team_id}{''.join(away_numbers)}{''.join(away_totals)}\n"
        f"{home_team_id}{''.join(home_numbers)}{''.join(home_totals)}"
    )


def add_filler_columns(list_to_pad):
    pad_length = 9 - len(list_to_pad)
    for i in range(pad_length):
        list_to_pad.append("  ")
    return list_to_pad


def get_linescore_inning_numbers(start, end):
    inning_numbers = []
    for i in range(start, end):
        pad_len = 1 if i < 10 else 0
        inning_numbers.append(f"{' '*pad_len}{i}")
    return inning_numbers


def get_innings_sorted(innings_unsorted):
    inning_ids = set([inning["inning_id"] for inning in innings_unsorted])
    inning_weights = {inning_id: get_inning_weight(inning_id) for inning_id in inning_ids}
    return sorted(innings_unsorted, key=lambda x: inning_weights[x["inning_id"]])


def get_inning_weight(inning_id):
    match = INNING_LABEL_REGEX.search(inning_id)
    if not match:
        return 0
    inn_half = match[1]
    inn_number = int(match[2])
    top_or_bottom = -1 if inn_half == "TOP" else 0
    return inn_number * 2 + top_or_bottom


def get_at_bat_details(at_bat, pitch_app_stats, bat_stats):
    pitch_stats = parse_pitch_app_stats(pitch_app_stats)
    (at_bats, details) = parse_bat_stats_for_game(bat_stats)
    inning = inning_number_to_string(at_bat["inning_id"][-5:])
    score = parse_score_for_at_bat(at_bat)
    outs_plural = "out" if at_bat["outs_before_play"] == 1 else "outs"
    outs = f"{at_bat['outs_before_play']} {outs_plural}"
    at_bat_result = parse_at_bat_result(at_bat)
    return (
        f"Pitcher..: {at_bat['pitcher_name']} ({pitch_stats})\n"
        f"Batter...: {at_bat['batter_name']} ({at_bats}, {details})\n"
        f"Score....: {score}\n"
        f"Inning...: {inning}, {outs}\n"
        f"On Base..: {at_bat['runners_on_base']}\n"
        f"Result...: {at_bat_result}"
    )


def parse_score_for_at_bat(at_bat):
    if not at_bat["pitchfx"]:
        return at_bat["score"]
    pfx = at_bat["pitchfx"][0]
    match = INNING_LABEL_REGEX.search(at_bat["inning_id"])
    if not match:
        return 0
    inn_half = match[1]
    batter_team_id = get_bbref_team_id(pfx["opponent_team_id_bb"])
    pitcher_team_id = get_bbref_team_id(pfx["pitcher_team_id_bb"])
    split = at_bat["score"].split("-")
    return (
        f"{batter_team_id} {split[0]} - {pitcher_team_id} {split[1]}"
        if inn_half == "TOP"
        else f"{pitcher_team_id} {split[1]} - {batter_team_id} {split[0]}"
    )


def parse_at_bat_result(at_bat):
    outcome = at_bat["play_description"]
    if at_bat["pitchfx"]:
        outcome = at_bat["pitchfx"][0]["des"]
    total_outs = at_bat["runs_outs_result"].count("O")
    total_runs = at_bat["runs_outs_result"].count("R")
    if not total_outs and not total_runs:
        return outcome
    outs_plural = "out" if total_outs == 1 else "outs"
    outs = f"{total_outs} {outs_plural}" if total_outs else ""
    runs_plural = "run" if total_runs == 1 else "runs"
    runs = f"{total_runs} {runs_plural} scored" if total_runs else ""
    separator = ", " if total_outs and total_runs else ""
    return f"{outcome} ({outs}{separator}{runs})"


def parse_bat_stats_for_game(bat_stats):
    if not bat_stats or not bat_stats["plate_appearances"]:
        return ("0/0", "")
    at_bats = f"{bat_stats['hits']}/{bat_stats['at_bats']}"
    stat_line = get_detailed_bat_stats(bat_stats)
    return (at_bats, stat_line)


def get_detailed_bat_stats(bat_stats):
    stats = homeruns = doubles = triples = stolen_bases = None
    stat_list = []
    if bat_stats["details"]:
        stats = parse_bat_stat_details(bat_stats["details"])
        homeruns = stats.pop("HR", None)
        triples = stats.pop("3B", None)
        doubles = stats.pop("2B", None)
        stolen_bases = stats.pop("SB", None)
    if homeruns:
        stat_list.append(homeruns)
    if triples:
        stat_list.append(triples)
    if doubles:
        stat_list.append(doubles)
    if bat_stats["rbis"]:
        rbis = bat_stats["rbis"] if bat_stats["rbis"] > 1 else ""
        stat_list.append(f"{rbis}RBI")
    if bat_stats["runs_scored"]:
        runs_scored = bat_stats["runs_scored"] if bat_stats["runs_scored"] > 1 else ""
        stat_list.append(f"{runs_scored}R")
    if stolen_bases:
        stat_list.append(stolen_bases)
    if bat_stats["strikeouts"]:
        strikeouts = bat_stats["strikeouts"] if bat_stats["strikeouts"] > 1 else ""
        stat_list.append(f"{strikeouts}K")
    if bat_stats["bases_on_balls"]:
        bases_on_balls = bat_stats["bases_on_balls"] if bat_stats["bases_on_balls"] > 1 else ""
        stat_list.append(f"{bases_on_balls}BB")
    if stats:
        stat_list.extend([remaining_stat for remaining_stat in stats.values()])
    return ", ".join(stat_list)


def parse_bat_stat_details(details):
    return {stat["stat"]: parse_single_bat_stat(stat) for stat in details}


def parse_single_bat_stat(stat):
    return f"{stat['count']}{stat['stat']}" if stat["count"] > 1 else stat["stat"]


def parse_bat_stats_to_date(bat_stats):
    return (
        (
            f"{format_stat_decimal(bat_stats['avg_to_date'])}/"
            f"{format_stat_decimal(bat_stats['obp_to_date'])}/"
            f"{format_stat_decimal(bat_stats['slg_to_date'])}/"
            f"{format_stat_decimal(bat_stats['ops_to_date'])}"
        )
        if bat_stats
        else ""
    )


def format_stat_decimal(stat_value):
    stat_with_leading_zero = f"{stat_value:0.3f}"
    return stat_with_leading_zero[1:]


def parse_pitch_app_stats(pitch_stats):
    if not pitch_stats or not pitch_stats["bbref_data"]:
        return ""
    bbref_stats = pitch_stats["bbref_data"]
    ip = bbref_stats["innings_pitched"]
    r = bbref_stats["runs"]
    er = bbref_stats["earned_runs"]
    runs = f"{er}ER" if er == r else f"{r}R, {er}ER"
    stat_list = []
    if bbref_stats["hits"]:
        h = bbref_stats["hits"] if bbref_stats["hits"] > 1 else ""
        stat_list.append(f"{h}H")
    if bbref_stats["strikeouts"]:
        k = bbref_stats["strikeouts"] if bbref_stats["strikeouts"] > 1 else ""
        stat_list.append(f"{k}K")
    if bbref_stats["bases_on_balls"]:
        bb = bbref_stats["bases_on_balls"] if bbref_stats["bases_on_balls"] > 1 else ""
        stat_list.append(f"{bb}BB")
    details = f", {', '.join(stat_list)}" if stat_list else ""
    game_score = f" (GS: {bbref_stats['game_score']})" if bbref_stats["game_score"] > 0 else ""
    return f"{ip} IP, {runs}{details}{game_score}"


def format_pitch_sequence_description(pitch_seq_desc):
    return [format_event_description(seq) for seq in pitch_seq_desc]


def format_event_description(event_desc):
    if (
        not event_desc[0]
        and not event_desc[2]
        and event_desc[1].startswith("(")
        and event_desc[1].endswith(")")
    ):
        event_desc[1] = event_desc[1][1:]
        event_desc[1] = event_desc[1][:-1]
    desc_lines = [wrap_text(s, max_len=AT_BAT_DESC_MAX_WIDTH) for s in event_desc[1].split("\n")]
    event_desc[1] = "\n".join(desc_lines)
    event_desc[1] = event_desc[1].replace("\n\n", "\n")
    return event_desc



"""Boxscore URLs for all games that took place on a single day."""
import json
from collections import ChainMap

from vigorish.util.list_helpers import as_dict_list, flatten_list2d


class BrooksGamesForDate:
    """Boxscore URLs for all games that took place on a single day."""

    dashboard_url = ""
    game_date = None
    game_date_str = ""
    game_count = ""
    games = []

    @property
    def game_id_dict(self):
        dict_list = [g.game_id_dict for g in self.games if not g.might_be_postponed]
        return dict(ChainMap(*dict_list))

    @property
    def all_pitch_app_ids_for_date(self):
        return sorted(flatten_list2d([g.all_pitch_app_ids for g in self.games]))

    def as_dict(self):
        """Convert daily boxscore URL values to a dictionary."""
        return dict(
            __brooks_games_for_date__=True,
            dashboard_url=self.dashboard_url,
            game_date_str=self.game_date_str,
            game_count=int(self.game_count),
            games=as_dict_list(self.games),
        )

    def as_json(self):
        """Convert daily game list to JSON."""
        return json.dumps(self.as_dict(), indent=2)

    def retrieve_game_info(self, bbref_game_id):
        game_match = [game for game in self.games if game.bbref_game_id == bbref_game_id]
        return game_match[0] if game_match else None



import subprocess
from pathlib import Path

from events import Events
from halo import Halo

from vigorish.cli.components import get_random_cli_color
from vigorish.config.database import Season
from vigorish.enums import DataSet, VigFile
from vigorish.status.update_status_bbref_boxscores import update_status_bbref_boxscore_list
from vigorish.status.update_status_bbref_games_for_date import update_bbref_games_for_date_list
from vigorish.status.update_status_brooks_games_for_date import (
    update_status_brooks_games_for_date_list,
)
from vigorish.status.update_status_brooks_pitch_logs import (
    update_status_brooks_pitch_logs_for_game_list,
)
from vigorish.status.update_status_brooks_pitchfx import update_status_brooks_pitchfx_log_list
from vigorish.tasks.base import Task
from vigorish.util.result import Result


class ImportScrapedDataTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.events = Events(
            (
                "error_occurred",
                "no_scraped_data_found",
                "remove_invalid_pitchfx_logs_start",
                "remove_invalid_pitchfx_logs_complete",
                "import_scraped_data_start",
                "import_scraped_data_complete",
            )
        )

    def execute(self, overwrite_existing=False):
        all_mlb_seasons = [season.year for season in Season.all_regular_seasons(self.db_session)]
        mlb_season_data_map = {
            year: self.local_folder_has_parsed_data_for_season(year) for year in all_mlb_seasons
        }
        if not any(has_data for has_data in mlb_season_data_map.values()):
            self.events.no_scraped_data_found()
            return Result.Ok()
        for year, data_exists in mlb_season_data_map.items():
            if not data_exists:
                continue
            subprocess.run(["clear"])
            result = self.remove_invalid_pitchfx_logs(year)
            if result.failure:
                return result
            result = self.update_all_data_sets(year, overwrite_existing)
            if result.failure:
                return result
        return Result.Ok()

    def local_folder_has_parsed_data_for_season(self, year):
        for data_set in DataSet:
            if data_set == DataSet.ALL:
                continue
            scraped_ids = self.scraped_data.get_scraped_ids_from_local_folder(
                VigFile.PARSED_JSON, data_set, year
            )
            if scraped_ids:
                return True
        return False

    def remove_invalid_pitchfx_logs(self, year):
        self.events.remove_invalid_pitchfx_logs_start(year)
        season = Season.find_by_year(self.db_session, year)
        if not season:
            return Result.Fail(f"No data has been scraped for the MLB {year} season.")
        valid_ids = []
        for game_date in season.get_date_range():
            games_for_date = self.scraped_data.get_brooks_games_for_date(game_date)
            if not games_for_date:
                continue
            valid_ids.extend(games_for_date.all_pitch_app_ids_for_date)
        folder_ids = self.scraped_data.get_scraped_ids_from_local_folder(
            VigFile.PARSED_JSON, DataSet.BROOKS_PITCHFX, season.year
        )
        invalid_ids = list(set(folder_ids) - set(valid_ids))
        if not invalid_ids:
            return Result.Ok()
        pfx_folderpath = self.scraped_data.get_local_folderpath(
            VigFile.PARSED_JSON, DataSet.BROOKS_PITCHFX, season.year
        )
        for pitch_app_id in invalid_ids:
            invalid_pitchfx_log = Path(pfx_folderpath).joinpath(f"{pitch_app_id}.json")
            if invalid_pitchfx_log.exists():
                invalid_pitchfx_log.unlink()
        self.events.remove_invalid_pitchfx_logs_complete(year)
        return Result.Ok()

    def update_all_data_sets(self, year, overwrite_existing=False):
        for data_set in DataSet:
            if data_set == DataSet.ALL:
                continue
            result = self.update_data_set(data_set, year, overwrite_existing)
            if result.failure:
                return result
        return Result.Ok()

    def update_data_set(self, data_set, year, overwrite_existing=False):
        self.events.import_scraped_data_start(data_set, year)
        scraped_ids = self.scraped_data.get_scraped_ids_from_local_folder(
            VigFile.PARSED_JSON, data_set, year
        )
        if not scraped_ids:
            return Result.Ok()
        spinner = Halo(spinner="dots3", color=get_random_cli_color())
        spinner.text = f"Updating {data_set} for MLB {year}..."
        spinner.start()
        if not overwrite_existing:
            result = self.scraped_data.get_scraped_ids_from_database(data_set, year)
            if result.failure:
                return result
            existing_scraped_ids = result.value
            scraped_ids = list(set(scraped_ids) - set(existing_scraped_ids))
        result = self.update_status_for_data_set(data_set, scraped_ids)
        if result.failure:
            self.events.error_occurred(result.error, data_set, year)
            return result
        self.db_session.commit()
        spinner.succeed(f"Successfully updated {data_set} for MLB {year}!")
        self.events.import_scraped_data_complete(data_set, year)
        return Result.Ok()

    def update_status_for_data_set(self, data_set, scraped_ids):
        update_status_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: update_status_brooks_games_for_date_list,
            DataSet.BROOKS_PITCH_LOGS: update_status_brooks_pitch_logs_for_game_list,
            DataSet.BROOKS_PITCHFX: update_status_brooks_pitchfx_log_list,
            DataSet.BBREF_GAMES_FOR_DATE: update_bbref_games_for_date_list,
            DataSet.BBREF_BOXSCORES: update_status_bbref_boxscore_list,
        }
        return update_status_dict[data_set](self.scraped_data, self.db_session, scraped_ids)



"""Boxscore URLs for all games that took place on a single day."""
import json


class BBRefGamesForDate:
    """Boxscore URLs for all games that took place on a single day."""

    dashboard_url = ""
    game_date = None
    game_date_str = ""
    game_count = ""
    boxscore_urls = []

    def as_dict(self):
        """Convert daily boxscore URL values to a dictionary."""
        return dict(
            __bbref_games_for_date__=True,
            dashboard_url=self.dashboard_url,
            game_date_str=self.game_date_str,
            game_count=int(self.game_count),
            boxscore_urls=self.boxscore_urls,
        )

    def as_json(self):
        """Convert daily game list to JSON."""
        return json.dumps(self.as_dict(), indent=2, sort_keys=True)



class BBRefTransactionDetail:
    player_id = 0
    player_id_br = ""
    draft_year = ""
    draft_round = ""
    draft_pick = ""
    old_team_id_br = 0
    new_team_id_br = 0

    def as_dict(self):
        dict = {
            "player_id": "{}".format(self.player_id),
            "player_id_br": "{}".format(self.player_id_br),
            "old_team_id_br": "{}".format(self.old_team_id_br),
            "new_team_id_br": "{}".format(self.new_team_id_br),
        }

        if self.draft_year and not self.draft_year.isspace():
            dict["draft_year"] = int(self.draft_year)
        if self.draft_round and not self.draft_round.isspace():
            dict["draft_round"] = int(self.draft_round)
        if self.draft_pick and not self.draft_pick.isspace():
            dict["draft_pick"] = int(self.draft_pick)
        return dict



import click
from dateutil import parser as date_parser

from vigorish.config.database import Season
from vigorish.util.regex import JOB_NAME_REGEX


class DateString(click.ParamType):
    name = "date-string"

    def convert(self, value, param, ctx):
        try:
            date = date_parser.parse(value)
            return date
        except Exception:
            error = (
                f'"{value}" could not be parsed as a valid date. You can '
                "use any format recognized by dateutil.parser. For example, "
                "all of the strings below are valid ways to represent the "
                "same date\n"
                '"2018-5-13" -or- "05/13/2018" -or- "May 13 2018"'
            )
            self.fail(error, param, ctx)


class MlbSeason(click.ParamType):
    name = "year-number"

    def convert(self, value, param, ctx):
        db_session = ctx.obj["db_session"]
        try:
            year = int(value)
        except Exception:
            error = f'Unable to parse "{value}" as an integer.'
            self.fail(error, param, ctx)

        try:
            season = Season.find_by_year(db_session, year)
            if season:
                return season.year
            valid_years = [s.year for s in Season.all_regular_seasons(db_session)]
            year_min = min(sorted(valid_years))
            year_max = max(sorted(valid_years))
            error = (
                f'"{value}" is not a valid value. Only MLB Seasons in '
                f"the range {year_min}-{year_max} are supported in this "
                "version of vig."
            )
            self.fail(error, param, ctx)
        except Exception:
            self.fail(error, param, ctx)


class JobName(click.ParamType):
    name = "job-name"

    def convert(self, value, param, ctx):
        if JOB_NAME_REGEX.match(value):
            return value
        error = (
            f"'{value}' contains one or more invalid characters. Job name must "
            "contain only letters, numbers, hyphen and underscore characters."
        )
        self.fail(error, param, ctx)



"""Biographical information for a single player."""
from sqlalchemy import Boolean, Column, DateTime, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import backref, relationship

from vigorish.config.database import Base
from vigorish.util.list_helpers import display_dict
from vigorish.util.pitch_calcs import calc_pitch_mix


class Player(Base):
    """Biographical information for a single player."""

    __tablename__ = "player"
    id = Column(Integer, primary_key=True)
    name_first = Column(String)
    name_last = Column(String)
    name_given = Column(String)
    bats = Column(String)
    throws = Column(String)
    weight = Column(Integer)
    height = Column(Integer)
    debut = Column(DateTime)
    birth_year = Column(Integer)
    birth_month = Column(Integer)
    birth_day = Column(Integer)
    birth_country = Column(String)
    birth_state = Column(String)
    birth_city = Column(String)
    bbref_id = Column(String, index=True)
    retro_id = Column(String, default="")
    mlb_id = Column(Integer, index=True)
    scraped_transactions = Column(Boolean, default=False)
    minor_league_player = Column(Boolean, default=False)
    missing_mlb_id = Column(Boolean, default=True)

    id_map = relationship("PlayerId", backref=backref("player", uselist=False))
    pitch_apps = relationship("PitchAppScrapeStatus", backref="player")
    pmix_all_view = relationship(
        "Pitch_Mix_All_View",
        uselist=False,
        primaryjoin="Player.id==Pitch_Mix_All_View.id",
        foreign_keys="Pitch_Mix_All_View.id",
    )
    pmix_right_view = relationship(
        "Pitch_Mix_Right_View",
        uselist=False,
        primaryjoin="Player.id==Pitch_Mix_Right_View.id",
        foreign_keys="Pitch_Mix_Right_View.id",
    )
    pmix_left_view = relationship(
        "Pitch_Mix_Left_View",
        uselist=False,
        primaryjoin="Player.id==Pitch_Mix_Left_View.id",
        foreign_keys="Pitch_Mix_Left_View.id",
    )

    def __repr__(self):
        return f"<Player name={self.name_first} {self.name_last}, bbref_id={self.bbref_id}>"

    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    def display(self):
        display_dict(self.as_dict())

    @hybrid_property
    def pitch_mix(self):
        return calc_pitch_mix(self.pmix_all_view.__dict__, self.pmix_all_view.total_pitches)

    @hybrid_property
    def pitch_mix_right(self):
        return calc_pitch_mix(self.pmix_right_view.__dict__, self.pmix_right_view.total_pitches)

    @hybrid_property
    def pitch_mix_left(self):
        return calc_pitch_mix(self.pmix_left_view.__dict__, self.pmix_left_view.total_pitches)

    @classmethod
    def find_by_bbref_id(cls, db_session, bbref_id):
        return db_session.query(cls).filter_by(bbref_id=bbref_id).first()

    @classmethod
    def find_by_mlb_id(cls, db_session, mlb_id):
        return db_session.query(cls).filter_by(mlb_id=mlb_id).first()



from halo import Halo

from vigorish.config.database import Season
from vigorish.constants import JOB_SPINNER_COLORS
from vigorish.enums import DataSet
from vigorish.status.update_status_bbref_boxscores import update_data_set_bbref_boxscores
from vigorish.status.update_status_bbref_games_for_date import update_data_set_bbref_games_for_date
from vigorish.status.update_status_brooks_games_for_date import (
    update_data_set_brooks_games_for_date,
)
from vigorish.status.update_status_brooks_pitch_logs import update_data_set_brooks_pitch_logs
from vigorish.status.update_status_brooks_pitchfx import update_data_set_brooks_pitchfx
from vigorish.util.result import Result


def update_status_for_mlb_season(db_session, scraped_data, year):
    spinner = Halo(color=JOB_SPINNER_COLORS[DataSet.BBREF_GAMES_FOR_DATE], spinner="dots3")
    spinner.text = f"Updating MLB {year}..."
    spinner.start()

    season = Season.find_by_year(db_session, year)
    if not season:
        spinner.stop_and_persist("😢", "Fail").clear()
        error = f"Error occurred retrieving season for year={year}"
        return Result.Fail(error)

    spinner.text = f"Updating MLB {year} bbref_games_for_date..."
    result = update_data_set_bbref_games_for_date(scraped_data, db_session, season)
    if result.failure:
        spinner.stop_and_persist("😢", "Fail").clear()
        return result
    db_session.commit()

    spinner.text = f"Updating MLB {year} brooks_games_for_date..."
    spinner.color = JOB_SPINNER_COLORS[DataSet.BROOKS_GAMES_FOR_DATE]
    result = update_data_set_brooks_games_for_date(scraped_data, db_session, season)
    if result.failure:
        spinner.stop_and_persist("😢", "Fail").clear()
        return result
    db_session.commit()

    spinner.text = f"Updating MLB {year} bbref_boxscores..."
    spinner.color = JOB_SPINNER_COLORS[DataSet.BBREF_BOXSCORES]
    result = update_data_set_bbref_boxscores(scraped_data, db_session, season)
    if result.failure:
        spinner.stop_and_persist("😢", "Fail").clear()
        return result
    db_session.commit()

    spinner.text = f"Updating MLB {year} brooks_pitch_logs..."
    spinner.color = JOB_SPINNER_COLORS[DataSet.BROOKS_PITCH_LOGS]
    result = update_data_set_brooks_pitch_logs(scraped_data, db_session, season)
    if result.failure:
        spinner.stop_and_persist("😢", "Fail").clear()
        return result
    db_session.commit()

    spinner.text = f"Updating MLB {year} brooks_pitchfx..."
    spinner.color = JOB_SPINNER_COLORS[DataSet.BROOKS_PITCHFX]
    result = update_data_set_brooks_pitchfx(scraped_data, db_session, season)
    if result.failure:
        spinner.stop_and_persist("😢", "Fail").clear()
        return result
    db_session.commit()

    spinner.stop_and_persist("😎", "Success").clear()
    return Result.Ok()



"""Pitch logs and pitchfx URLs for a single game."""
import json

from vigorish.util.list_helpers import as_dict_list
from vigorish.util.string_helpers import validate_brooks_game_id


class BrooksPitchLogsForGame:
    """Pitch logs and pitchfx URLs for a single game."""

    bb_game_id = ""
    bbref_game_id = ""
    pitch_log_count = ""
    pitch_logs = []

    @property
    def game_id_dict(self):
        return {f"{self.bbref_game_id}": f"{self.bb_game_id}"}

    @property
    def game_date(self):
        result = validate_brooks_game_id(self.bb_game_id)
        if result.failure:
            return None
        game_dict = result.value
        return game_dict["game_date"]

    @property
    def total_pitch_count(self):
        return sum(pitch_log.total_pitch_count for pitch_log in self.pitch_logs)

    @property
    def upload_id(self):
        return self.bbref_game_id

    @property
    def pitch_app_ids(self):
        return [pitch_log.pitch_app_id for pitch_log in self.pitch_logs]

    @property
    def pitch_app_ids_scraped_all_data(self):
        return [
            pitch_log.pitch_app_id for pitch_log in self.pitch_logs if pitch_log.parsed_all_info
        ]

    @property
    def pitch_app_ids_no_data(self):
        return [
            pitch_log.pitch_app_id for pitch_log in self.pitch_logs if not pitch_log.parsed_all_info
        ]

    def as_dict(self):
        """Convert pitch logs for game to a dictionary."""
        return dict(
            __brooks_pitch_logs_for_game__=True,
            bb_game_id=self.bb_game_id,
            bbref_game_id=self.bbref_game_id,
            pitch_log_count=int(self.pitch_log_count),
            pitch_logs=as_dict_list(self.pitch_logs),
        )

    def as_json(self):
        """Convert pitch logs for game to JSON."""
        return json.dumps(self.as_dict(), indent=2)



from dataclasses import dataclass
from typing import Any

from vigorish.util.list_helpers import as_dict_list


@dataclass
class BBRefBoxscoreTeamData:
    starting_lineup: Any
    batting_stats: Any
    pitching_stats: Any
    team_id_br: str = ""
    total_wins_before_game: int = 0
    total_losses_before_game: int = 0
    total_runs_scored_by_team: int = 0
    total_runs_scored_by_opponent: int = 0
    total_hits_by_team: int = 0
    total_hits_by_opponent: int = 0
    total_errors_by_team: int = 0
    total_errors_by_opponent: int = 0

    def as_dict(self):
        return dict(
            __bbref_boxscore_team_data__=True,
            team_id_br=self.team_id_br,
            total_wins_before_game=self.total_wins_before_game,
            total_losses_before_game=self.total_losses_before_game,
            total_runs_scored_by_team=self.total_runs_scored_by_team,
            total_runs_scored_by_opponent=self.total_runs_scored_by_opponent,
            total_hits_by_team=self.total_hits_by_team,
            total_hits_by_opponent=self.total_hits_by_opponent,
            total_errors_by_team=self.total_errors_by_team,
            total_errors_by_opponent=self.total_errors_by_opponent,
            starting_lineup=as_dict_list(self.starting_lineup),
            batting_stats=as_dict_list(self.batting_stats),
            pitching_stats=as_dict_list(self.pitching_stats),
        )



"""Enum definitions."""
from enum import auto, Enum, IntEnum

from aenum import auto as auto_flag
from aenum import IntFlag


class VigFile(IntFlag):
    """File types that are scraped, created and stored by vigorish."""

    SCRAPED_HTML = auto_flag()
    PARSED_JSON = auto_flag()
    COMBINED_GAME_DATA = auto_flag()
    PATCH_LIST = auto_flag()
    ALL = SCRAPED_HTML | PARSED_JSON | COMBINED_GAME_DATA | PATCH_LIST

    def __str__(self):
        return self.name


class DataSet(IntFlag):
    """MLB data sets."""

    BBREF_GAMES_FOR_DATE = auto_flag()
    BROOKS_GAMES_FOR_DATE = auto_flag()
    BBREF_BOXSCORES = auto_flag()
    BROOKS_PITCH_LOGS = auto_flag()
    BROOKS_PITCHFX = auto_flag()
    ALL = (
        BBREF_GAMES_FOR_DATE
        | BROOKS_GAMES_FOR_DATE
        | BBREF_BOXSCORES
        | BROOKS_PITCH_LOGS
        | BROOKS_PITCHFX
    )

    def __str__(self):
        return self.name


class ConfigType(Enum):
    """Data types for configuration settings."""

    STRING = auto()
    NUMERIC = auto()
    PATH = auto()
    ENUM = auto()
    NONE = auto()

    def __str__(self):
        return self.name


class ScrapeCondition(Enum):
    """Allowed values for SCRAPE_CONDITION config setting."""

    ONLY_MISSING_DATA = auto()
    ALWAYS = auto()
    NEVER = auto()

    def __str__(self):
        return self.name


class HtmlStorageOption(Enum):
    """Allowed values for HTML_STORAGE config setting."""

    NONE = auto()
    LOCAL_FOLDER = auto()
    S3_BUCKET = auto()
    BOTH = auto()

    def __str__(self):
        return self.name


class JsonStorageOption(Enum):
    """Allowed values for JSON_STORAGE config setting."""

    LOCAL_FOLDER = auto()
    S3_BUCKET = auto()
    BOTH = auto()

    def __str__(self):
        return self.name


class CombinedDataStorageOption(Enum):
    """Allowed values for COMBINED_DATA_STORAGE config setting."""

    LOCAL_FOLDER = auto()
    S3_BUCKET = auto()
    BOTH = auto()

    def __str__(self):
        return self.name


class StatusReport(Enum):
    """The type of status report (if any) to display."""

    NONE = 0
    SEASON_SUMMARY = 1
    DATE_SUMMARY_MISSING_DATA = 2
    DATE_SUMMARY_ALL_DATES = 3
    DATE_DETAIL_MISSING_DATA = 4
    DATE_DETAIL_ALL_DATES = 5
    DATE_DETAIL_MISSING_PITCHFX = 6
    SINGLE_DATE_WITH_GAME_STATUS = 7

    def __str__(self):
        return self.name


class FileTask(Enum):
    """Generic file actions that are mapped to LocalFileTask and S3FileTask file actions."""

    STORE_FILE = auto()
    GET_FILE = auto()
    REMOVE_FILE = auto()

    def __str__(self):
        return self.name


class S3FileTask(Enum):
    """Type of action to perform on an object in an S3 bucket."""

    UPLOAD = auto()
    DOWNLOAD = auto()
    DELETE = auto()
    RENAME = auto()

    def __str__(self):
        return self.name


class LocalFileTask(Enum):
    """Type of action to perform on a local file."""

    WRITE_FILE = auto()
    READ_FILE = auto()
    DELETE_FILE = auto()
    DECODE_JSON = auto()

    def __str__(self):
        return self.name


class JobStatus(IntEnum):
    """Status of a scrape job created by a user."""

    INCOMPLETE = auto()
    ERROR = auto()
    COMPLETE = auto()

    def __str__(self):
        return self.name


class DefensePosition(IntEnum):
    """Defensive positions and position numbers used when scoring a game."""

    NONE = 0
    PITCHER = 1
    CATCHER = 2
    FIRST_BASE = 3
    SECOND_BASE = 4
    THIRD_BASE = 5
    SHORT_STOP = 6
    LEFT_FIELD = 7
    CENTER_FIELD = 8
    RIGHT_FIELD = 9
    DH = 10

    def __str__(self):
        abbrev_dict = {
            "NONE": "BN",
            "PITCHER": "P",
            "CATCHER": "C",
            "FIRST_BASE": "1B",
            "SECOND_BASE": "2B",
            "THIRD_BASE": "3B",
            "SHORT_STOP": "SS",
            "LEFT_FIELD": "LF",
            "CENTER_FIELD": "CF",
            "RIGHT_FIELD": "RF",
        }
        return abbrev_dict.get(self.name, self.name)

    @classmethod
    def from_abbrev(cls, abbrev):
        abbrev_dict = {
            "P": cls.PITCHER,
            "C": cls.CATCHER,
            "1B": cls.FIRST_BASE,
            "2B": cls.SECOND_BASE,
            "3B": cls.THIRD_BASE,
            "SS": cls.SHORT_STOP,
            "LF": cls.LEFT_FIELD,
            "CF": cls.CENTER_FIELD,
            "RF": cls.RIGHT_FIELD,
            "DH": cls.DH,
        }
        return abbrev_dict.get(abbrev, cls.NONE)


class SeasonType(Enum):
    """Identifies games, stats, etc as Spring Training, Regular, or Post-Season."""

    PRE_SEASON = auto()
    REGULAR_SEASON = auto()
    POST_SEASON = auto()
    NONE = auto()

    def __str__(self):
        season_type_dict = {
            "PRE_SEASON": "Spring Training",
            "REGULAR_SEASON": "Regular Season",
            "POST_SEASON": "Post-Season",
        }
        return season_type_dict.get(self.name, self.name)


class PlayByPlayEvent(Enum):
    AT_BAT = auto()
    SUBSTITUTION = auto()
    MISC = auto()

    def __str__(self):
        return self.name


class AuditError(Enum):
    FAILED_TO_COMBINE = auto()
    PITCHFX_ERROR = auto()
    INVALID_PITCHFX_DATA = auto()

    def __str__(self):
        error_type_dict = {
            "FAILED_TO_COMBINE": "Failed to combine scraped data",
            "PITCHFX_ERROR": "PitchFX error",
            "INVALID_PITCHFX_DATA": "Invalid PitchFX data",
        }
        return error_type_dict.get(self.name, self.name)


class AuditTask(IntEnum):
    GET_SCRAPED_DATA = 1
    GET_PLAY_BY_PLAY_DATA = 2
    GET_PITCHFX_DATA = 3
    COMBINE_PBP_AND_PFX_DATA = 4
    UPDATE_BOXSCORE = 5

    def __str__(self):
        return self.name


class PatchType(Enum):
    CHANGE_BBREF_GAME_ID = auto()
    CHANGE_PITCH_SEQUENCE = auto()
    CHANGE_BATTER_ID = auto()
    CHANGE_PITCHER_ID = auto()

    def __str__(self):
        return self.name


class SyncDirection(Enum):
    UP_TO_S3 = auto()
    DOWN_TO_LOCAL = auto()

    def __str__(self):
        return self.name



from vigorish.config.database import DateScrapeStatus, GameScrapeStatus, Season
from vigorish.enums import DataSet
from vigorish.util.dt_format_strings import DATE_ONLY, DATE_ONLY_TABLE_ID
from vigorish.util.result import Result


def update_brooks_games_for_date_single_date(db_session, season, games_for_date):
    game_date = games_for_date.game_date
    result = update_date_status_records(db_session, games_for_date, game_date)
    if result.failure:
        return result
    result = update_game_status_records(db_session, season, games_for_date.games)
    if result.failure:
        return result
    return Result.Ok()


def update_data_set_brooks_games_for_date(scraped_data, db_session, season):
    result = scraped_data.get_all_brooks_dates_scraped(season.year)
    if result.failure:
        return result
    scraped_brooks_dates = result.value
    unscraped_brooks_dates = DateScrapeStatus.get_all_brooks_unscraped_dates_for_season(
        db_session, season.id
    )
    new_brooks_dates = set(scraped_brooks_dates) & set(unscraped_brooks_dates)
    if not new_brooks_dates:
        return Result.Ok()
    result = update_status_brooks_games_for_date_list(scraped_data, db_session, new_brooks_dates)
    if result.failure:
        return result
    new_brooks_games = result.value
    return update_game_status_records(db_session, season, new_brooks_games)


def update_status_brooks_games_for_date_list(scraped_data, db_session, scraped_brooks_dates):
    season = None
    for game_date in scraped_brooks_dates:
        if not season:
            season = Season.find_by_year(db_session, game_date.year)
        games_for_date = scraped_data.get_brooks_games_for_date(game_date)
        if not games_for_date:
            date_str = game_date.strftime(DATE_ONLY_TABLE_ID)
            error = f"Failed to retrieve {DataSet.BROOKS_GAMES_FOR_DATE} (URL ID: {date_str})"
            return Result.Fail(error)
        result = update_date_status_records(db_session, games_for_date, game_date)
        if result.failure:
            return result
        result = update_game_status_records(db_session, season, games_for_date.games)
        if result.failure:
            return result
        db_session.commit()
    return Result.Ok()


def update_date_status_records(db_session, games_for_date, game_date):
    try:
        date_id = game_date.strftime(DATE_ONLY_TABLE_ID)
        date_status = db_session.query(DateScrapeStatus).get(date_id)
        if not date_status:
            date_str = games_for_date.game_date.strftime(DATE_ONLY)
            error = f"scrape_status_date does not contain an entry for date: {date_str}"
            return Result.Fail(error)
        date_status.scraped_daily_dash_brooks = 1
        date_status.game_count_brooks = games_for_date.game_count
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")


def update_game_status_records(db_session, season, new_brooks_games):
    for brooks_game_info in new_brooks_games:
        if not brooks_game_info.pitcher_appearance_count:
            continue
        try:
            bbref_game_id = brooks_game_info.bbref_game_id
            game_date = brooks_game_info.game_date
            date_status = DateScrapeStatus.find_by_date(db_session, game_date)
            game_status = GameScrapeStatus.find_by_bbref_game_id(db_session, bbref_game_id)
            if not game_status:
                error = f"scrape_status_game does not contain a record for game: {bbref_game_id}"
                return Result.Fail(error)
            game_status.game_date = game_date
            game_status.bbref_game_id = bbref_game_id
            game_status.bb_game_id = brooks_game_info.bb_game_id
            game_status.game_time_hour = brooks_game_info.game_time_hour
            game_status.game_time_minute = brooks_game_info.game_time_minute
            game_status.game_time_zone = brooks_game_info.time_zone_name
            game_status.pitch_app_count_brooks = brooks_game_info.pitcher_appearance_count
            game_status.scrape_status_date_id = date_status.id
            game_status.season_id = season.id
            db_session.add(game_status)
        except Exception as e:
            return Result.Fail(f"Error: {repr(e)}")
    return Result.Ok()



from vigorish.config.database import PitchAppScrapeStatus
from vigorish.enums import DataSet
from vigorish.util.result import Result


def update_data_set_brooks_pitchfx(scraped_data, db_session, season):
    scraped_ids = scraped_data.get_all_scraped_pitchfx_pitch_app_ids(season.year)
    unscraped_ids = PitchAppScrapeStatus.get_all_unscraped_pitch_app_ids_for_season(
        db_session, season.id
    )
    new_pitch_app_ids = set(scraped_ids) & set(unscraped_ids)
    if not new_pitch_app_ids:
        return Result.Ok()
    result = update_status_brooks_pitchfx_log_list(scraped_data, db_session, new_pitch_app_ids)
    if result.failure:
        return result
    return Result.Ok()


def update_status_brooks_pitchfx_log_list(scraped_data, db_session, new_pitch_app_ids):
    for pitch_app_id in new_pitch_app_ids:
        pitchfx_log = scraped_data.get_brooks_pitchfx_log(pitch_app_id)
        if not pitchfx_log:
            error = f"Failed to retrieve {DataSet.BROOKS_PITCHFX} (URL ID: {pitch_app_id})"
            return Result.Fail(error)
        result = update_pitch_appearance_status_records(db_session, pitchfx_log)
        if result.failure:
            return result
        db_session.commit()
    return Result.Ok()


def update_pitch_appearance_status_records(db_session, pitchfx_log):
    try:
        pitch_app_status = PitchAppScrapeStatus.find_by_pitch_app_id(
            db_session, pitchfx_log.pitch_app_id
        )
        if not pitch_app_status:
            error = (
                f"scrape_status_pitch_app does not contain an "
                f"entry for pitch_app_id: {pitchfx_log.pitch_app_id}"
            )
            return Result.Fail(error)
        pitch_app_status.scraped_pitchfx = 1
        pitch_app_status.pitch_count_pitchfx = pitchfx_log.total_pitch_count
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")



from sqlalchemy import select

from vigorish.config.database import Pitch_Mix_By_Year_View, Player, Season
from vigorish.util.pitch_calcs import calc_pitch_mix


class AllPlayerData:
    def __init__(self, app, mlb_id):
        self.app = app
        self.db_engine = app["db_engine"]
        self.db_session = app["db_session"]
        self.mlb_id = mlb_id
        self.player = Player.find_by_mlb_id(self.db_session, self.mlb_id)

    def pitch_mix_by_year(self):
        s = select([Pitch_Mix_By_Year_View]).where(Pitch_Mix_By_Year_View.id == self.player.id)
        results = self.db_engine.execute(s).fetchall()
        pitch_mix = {}
        for d in [dict(row) for row in results]:
            season = self.db_session.query(Season).get(d["season_id"])
            pitch_mix[season.year] = calc_pitch_mix(d, d["total_pitches"])
        return pitch_mix

    def pitch_mix(self):
        return self.player.pitch_mix

    def pitch_mix_right(self):
        return self.player.pitch_mix_right

    def pitch_mix_left(self):
        return self.player.pitch_mix_left



"""Update bbref_player_id_map.json file."""
import subprocess

from halo import Halo

from vigorish.cli.components import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_message,
    user_options_prompt,
    yes_no_prompt,
)
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJI_DICT, MENU_NUMBERS
from vigorish.tasks.import_scraped_data import ImportScrapedDataTask
from vigorish.util.result import Result

IMPORT_DATA_MESSAGE = (
    "This task uses the value of the JSON_LOCAL_FOLDER_PATH config setting to locate scraped "
    "data in the local file system from all data sets and all MLB seasons. After the search is "
    "complete, the scraped data is used to update the database."
)
IMPORT_DATA_PROMPT = "Select YES to run this task\nSelect NO to return to the previous menu"
OVERWRITE_DATA_MESSAGE = (
    'By default, data that is already considered "scraped" within the database will not be '
    "updated with the data found in the local file system."
)
OVERWRITE_DATA_PROMPT = (
    "Select KEEP EXISTING DATA to use the default behavior and only update data which is not "
    'already marked as "scraped" in the database\n'
    "Select OVERWRITE EXISTING DATA to update all data in the database with the values found in "
    "the local file system"
)


class ImportScrapedData(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.import_scraped_data = ImportScrapedDataTask(self.app)
        self.spinner = Halo(color=get_random_cli_color(), spinner=get_random_dots_spinner())
        self.menu_item_text = "Import Scraped Data from Local Folders"
        self.menu_item_emoji = EMOJI_DICT.get("HONEY_POT")

    def launch(self):
        if not self.prompt_user_import_data():
            return Result.Ok(True)
        self.subscribe_to_events()
        result = self.prompt_user_overwrite_data()
        if result.failure:
            return Result.Ok(True)
        overwrite_existing_data = True if result.value == "OVERWRITE" else False
        result = self.import_scraped_data.execute(overwrite_existing_data)
        self.unsubscribe_from_events()
        if result.failure:
            return result
        return Result.Ok(True)

    def prompt_user_import_data(self):
        subprocess.run(["clear"])
        print_message(IMPORT_DATA_MESSAGE, fg="bright_yellow", bold=True)
        return yes_no_prompt(IMPORT_DATA_PROMPT, wrap=False)

    def prompt_user_overwrite_data(self):
        subprocess.run(["clear"])
        print_message(OVERWRITE_DATA_MESSAGE, fg="bright_yellow", bold=True)
        choices = {
            f"{MENU_NUMBERS.get(1)}  KEEP EXISTING DATA": "KEEP",
            f"{MENU_NUMBERS.get(2)}  OVERWRITE EXISTING DATA": "OVERWRITE",
            f"{EMOJI_DICT.get('BACK')} Return to Admin/Tasks Menu": None,
        }
        return user_options_prompt(choices, OVERWRITE_DATA_PROMPT, clear_screen=False)

    def error_occurred(self, error_message, data_set, year):
        self.spinner.fail(f"Error occurred while updating {data_set} for MLB {year}")

    def subscribe_to_events(self):
        self.import_scraped_data.events.error_occurred += self.error_occurred

    def unsubscribe_from_events(self):
        self.import_scraped_data.events.error_occurred -= self.error_occurred



from vigorish.config.database import (
    DateScrapeStatus,
    GameScrapeStatus,
    PitchAppScrapeStatus,
    Season,
)
from vigorish.enums import DataSet
from vigorish.util.result import Result


def update_data_set_brooks_pitch_logs(scraped_data, db_session, season):
    result = scraped_data.get_all_scraped_brooks_game_ids(season.year)
    if result.failure:
        return result
    scraped_brooks_game_ids = result.value
    unscraped_brooks_game_ids = GameScrapeStatus.get_all_unscraped_brooks_game_ids_for_season(
        db_session, season.id
    )
    new_brooks_game_ids = set(scraped_brooks_game_ids) & set(unscraped_brooks_game_ids)
    if not new_brooks_game_ids:
        return Result.Ok()
    result = update_status_brooks_pitch_logs_for_game_list(
        scraped_data, db_session, season, new_brooks_game_ids
    )
    if result.failure:
        return result
    return Result.Ok()


def update_status_brooks_pitch_logs_for_game_list(scraped_data, db_session, new_brooks_game_ids):
    for bb_game_id in new_brooks_game_ids:
        pitch_logs_for_game = scraped_data.get_brooks_pitch_logs_for_game(bb_game_id)
        if not pitch_logs_for_game:
            error = f"Failed to retrieve {DataSet.BROOKS_PITCH_LOGS} (URL ID: {bb_game_id})"
            return Result.Fail(error)
        result = update_game_status_records(db_session, pitch_logs_for_game)
        if result.failure:
            return result
        result = create_pitch_appearance_status_records(db_session, pitch_logs_for_game)
        if result.failure:
            return result
        db_session.commit()
    return Result.Ok()


def update_status_brooks_pitch_logs_for_game(db_session, pitch_logs_for_game):
    result = update_game_status_records(db_session, pitch_logs_for_game)
    if result.failure:
        return result
    return create_pitch_appearance_status_records(db_session, pitch_logs_for_game)


def update_game_status_records(db_session, pitch_logs_for_game):
    try:
        bb_game_id = pitch_logs_for_game.bb_game_id
        game_status = GameScrapeStatus.find_by_bb_game_id(db_session, bb_game_id)
        if not game_status:
            error = f"scrape_status_game does not contain an entry for bb_game_id: {bb_game_id}"
            return Result.Fail(error)
        game_status.scraped_brooks_pitch_logs = 1
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")


def create_pitch_appearance_status_records(db_session, pitch_logs_for_game):
    season = None
    bbref_game_id = pitch_logs_for_game.bbref_game_id
    all_pitch_app_ids = PitchAppScrapeStatus.get_all_pitch_app_ids_for_game(
        db_session, bbref_game_id
    )
    scraped_pitch_app_ids = [plog.pitch_app_id for plog in pitch_logs_for_game.pitch_logs]
    update_pitch_app_ids = set(scraped_pitch_app_ids).difference(all_pitch_app_ids)
    for pitch_log in pitch_logs_for_game.pitch_logs:
        if pitch_log.pitch_app_id not in update_pitch_app_ids:
            continue
        if not season:
            season = Season.find_by_year(db_session, pitch_log.game_date.year)
        try:
            date_status = DateScrapeStatus.find_by_date(db_session, pitch_log.game_date)
            game_status = GameScrapeStatus.find_by_bb_game_id(db_session, pitch_log.bb_game_id)
            pitch_app_status = PitchAppScrapeStatus()
            pitch_app_status.pitcher_id_mlb = pitch_log.pitcher_id_mlb
            pitch_app_status.pitch_app_id = pitch_log.pitch_app_id
            pitch_app_status.bbref_game_id = pitch_log.bbref_game_id
            pitch_app_status.bb_game_id = pitch_log.bb_game_id
            pitch_app_status.pitch_count_pitch_log = pitch_log.total_pitch_count
            pitch_app_status.scrape_status_game_id = game_status.id
            pitch_app_status.scrape_status_date_id = date_status.id
            pitch_app_status.season_id = season.id
            if not pitch_log.parsed_all_info:
                pitch_app_status.scraped_pitchfx = 1
                pitch_app_status.no_pitchfx_data = 1
            db_session.add(pitch_app_status)
        except Exception as e:
            return Result.Fail(f"Error: {repr(e)}")
    return Result.Ok()



"""Menu that allows the user to view and modify all settings in vig.config.json."""
from vigorish.cli.menu import Menu
from vigorish.cli.menu_items.admin_tasks.add_pitchfx_to_database import AddPitchFxToDatabase
from vigorish.cli.menu_items.admin_tasks.backup_database import BackupDatabase
from vigorish.cli.menu_items.admin_tasks.calculate_avg_pitch_times import (
    CalculatePitchTimes,
)
from vigorish.cli.menu_items.admin_tasks.import_scraped_data import ImportScrapedData
from vigorish.cli.menu_items.admin_tasks.npm_install_update import NpmInstallUpdate
from vigorish.cli.menu_items.admin_tasks.restore_database import RestoreDatabase
from vigorish.cli.menu_items.admin_tasks.sync_scraped_data import SyncScrapedData
from vigorish.cli.menu_items.admin_tasks.update_player_id_map import UpdatePlayerIdMap
from vigorish.cli.menu_items.return_to_parent import ReturnToParent
from vigorish.cli.menu_items.setup_db import SetupDatabase
from vigorish.config.database import db_setup_complete
from vigorish.constants import EMOJI_DICT
from vigorish.util.result import Result
from vigorish.util.sys_helpers import node_is_installed


class AdminTasksMenu(Menu):
    def __init__(self, app, audit_report):
        # TODO: Add Refresh Game Data menu option
        # TODO: Add Admin Task menu item to edit season dates and add new seasons
        super().__init__(app)
        self.audit_report = audit_report
        self.menu_text = "Select a task from the list:"
        self.menu_item_text = "Tasks/Admin"
        self.menu_item_emoji = EMOJI_DICT.get("PAGER", "")

    def populate_menu_items(self):
        self.menu_items.clear()
        if node_is_installed():
            self.menu_items.append(NpmInstallUpdate(self.app))
        self.menu_items.append(UpdatePlayerIdMap(self.app))
        self.menu_items.append(SyncScrapedData(self.app))
        if db_setup_complete(self.db_engine, self.db_session):
            self.menu_items.append(ImportScrapedData(self.app))
            self.menu_items.append(AddPitchFxToDatabase(self.app, self.audit_report))
            self.menu_items.append(BackupDatabase(self.app))
            self.menu_items.append(RestoreDatabase(self.app))
            self.menu_items.append(CalculatePitchTimes(self.app))
            self.menu_items.append(SetupDatabase(self.app))
        self.menu_items.append(ReturnToParent(self.app, "Return to Main Menu"))
        # self.menu_items.insert(0, ReturnToParentMenuItem(self.app, "Return to Settings/Admin"))

    def prompt_user_for_menu_selection(self):
        menu = self._get_bullet_menu()
        menu.pos = self.selected_menu_item_pos
        self.selected_menu_item_text = menu.launch()
        result = self.selected_menu_item.launch()
        exit_menu = self.selected_menu_item.exit_menu
        return Result.Ok(exit_menu) if result.success else result



# flake8: noqa
from vigorish.cli.components.csv_viewer import CsvFileTableViewer
from vigorish.cli.components.data_set_check import DataSetCheck
from vigorish.cli.components.date_input import DateInput
from vigorish.cli.components.dict_viewer import DictListTableViewer
from vigorish.cli.components.job_name_input import JobNameInput
from vigorish.cli.components.models import DisplayPage, DisplayTable
from vigorish.cli.components.page_viewer import PageViewer
from vigorish.cli.components.prompts import *
from vigorish.cli.components.table_viewer import TableViewer
from vigorish.cli.components.util import *



"""ABC for menu items that execute a command."""
from abc import ABC, abstractmethod

from bullet import colors


class MenuItem(ABC):
    _menu_item_text = ""
    menu_item_emoji = ""
    background_color = colors.background["default"]
    background_on_switch = colors.background["default"]
    word_color = colors.foreground["default"]
    word_on_switch = colors.bright(colors.foreground["cyan"])
    exit_menu = False

    def __init__(self, app):
        self.app = app
        self.dotenv = app["dotenv"]
        self.config = app["config"]
        self.db_engine = app["db_engine"]
        self.db_session = app["db_session"]
        self.scraped_data = app["scraped_data"]

    @property
    def menu_item_text(self):
        return f"{self.menu_item_emoji} {self._menu_item_text}"

    @menu_item_text.setter
    def menu_item_text(self, menu_item_text):
        self._menu_item_text = menu_item_text

    @abstractmethod
    def launch(self):
        pass



import subprocess

from getch import pause
from halo import Halo

from vigorish.cli.components.prompts import user_options_prompt, yes_no_prompt
from vigorish.cli.components.util import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_message,
)
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJI_DICT, MENU_NUMBERS
from vigorish.tasks.add_pitchfx_to_database import AddPitchFxToDatabase as AddPitchFxToDatabaseTask
from vigorish.util.result import Result


class AddPitchFxToDatabase(MenuItem):
    def __init__(self, app, audit_report):
        super().__init__(app)
        self.audit_report = audit_report
        self.add_pfx_to_db = AddPitchFxToDatabaseTask(self.app)
        self.menu_item_text = "Add PitchFx Data to Database"
        self.menu_item_emoji = EMOJI_DICT.get("BASEBALL")
        self.exit_menu = False
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.game_ids = []

    @property
    def total_games(self):
        return len(self.game_ids)

    def launch(self):
        if not self.prompt_user_run_task():
            return Result.Ok(True)
        result = self.select_season_prompt()
        if result.failure:
            return Result.Ok()
        year = result.value
        self.initialize_spinner()
        self.subscribe_to_events()
        result = self.add_pfx_to_db.execute(self.audit_report, year)
        self.spinner.stop()
        self.unsubscribe_from_events()
        if result.failure:
            return result
        subprocess.run(["clear"])
        print_message(f"Added PitchFX data for {len(self.game_ids)} games to database!")
        pause(message="\nPress any key to continue...")
        return Result.Ok()

    def prompt_user_run_task(self):
        task_description = [
            "This task identifies all games that fulfill the two requirements below:\n",
            "1. All data sets have been scraped",
            "2. Data was successfully combined (does not need to be error free)\n",
            "PitchFX data for all games that meet these requirements will be added to the "
            "database.",
        ]
        subprocess.run(["clear"])
        for line in task_description:
            print_message(line, fg="bright_yellow")
        return yes_no_prompt("Would you like to run this task?", wrap=False)

    def select_season_prompt(self):
        prompt = "Select an MLB season from the list below:"
        choices = {f"{MENU_NUMBERS.get(1)}  ALL": "ALL"}
        for num, (year, results) in enumerate(self.audit_report.items(), start=2):
            if results["successful"]:
                choices[f"{MENU_NUMBERS.get(num, str(num))}  {year}"] = year
        choices[f"{EMOJI_DICT.get('BACK')} Return to Previous Menu"] = None
        result = user_options_prompt(choices, prompt)
        if result.failure:
            return result
        year = result.value if isinstance(result.value, int) else None
        return Result.Ok(year)

    def initialize_spinner(self):
        subprocess.run(["clear"])
        self.spinner.text = "Preparing to import PitchFx data..."
        self.spinner.start()

    def add_pitchfx_to_db_start(self, year, game_ids):
        self.game_ids = game_ids
        self.spinner.text = self.get_progress_text(0, year, game_ids[0])

    def add_pitchfx_to_db_progress(self, num_complete, year, game_id):
        self.spinner.text = self.get_progress_text(num_complete, year, game_id)

    def get_progress_text(self, num_complete, year, game_id):
        percent = num_complete / float(self.total_games)
        return f"Adding PitchFx for MLB {year} to Database... (Game ID: {game_id}) {percent:.0%}..."

    def subscribe_to_events(self):
        self.add_pfx_to_db.events.add_pitchfx_to_db_start += self.add_pitchfx_to_db_start
        self.add_pfx_to_db.events.add_pitchfx_to_db_progress += self.add_pitchfx_to_db_progress

    def unsubscribe_from_events(self):
        self.add_pfx_to_db.events.add_pitchfx_to_db_start -= self.add_pitchfx_to_db_start
        self.add_pfx_to_db.events.add_pitchfx_to_db_progress -= self.add_pitchfx_to_db_progress



from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from vigorish.config.config_file import ConfigFile
from vigorish.config.database import get_db_url
from vigorish.config.dotenv_file import DotEnvFile
from vigorish.data.scraped_data import ScrapedData


def create_app(db_url=None):
    dotenv = DotEnvFile()
    config = ConfigFile()
    if not db_url:
        db_url = get_db_url()
    db_engine = create_engine(db_url)
    session_maker = sessionmaker(bind=db_engine)
    db_session = session_maker()
    scraped_data = ScrapedData(db_engine, db_session, config)
    return {
        "dotenv": dotenv,
        "config": config,
        "db_engine": db_engine,
        "db_session": db_session,
        "scraped_data": scraped_data,
    }



import os
from datetime import datetime, timezone
from pathlib import Path
from zipfile import ZIP_DEFLATED, ZipFile

from events import Events

from vigorish.config.database import (
    DateScrapeStatus,
    GameScrapeStatus,
    get_total_number_of_rows,
    PitchAppScrapeStatus,
    PitchFx,
)
from vigorish.enums import DataSet
from vigorish.tasks.base import Task
from vigorish.util.dt_format_strings import FILE_TIMESTAMP
from vigorish.util.numeric_helpers import ONE_PERCENT
from vigorish.util.result import Result


class BackupDatabaseTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.events = Events(
            (
                "backup_database_start",
                "backup_table_start",
                "backup_table_progress",
                "backup_table_complete",
                "create_zip_file_start",
                "create_zip_file_complete",
                "backup_database_complete",
            )
        )

    def execute(self):
        csv_map = self.get_table_filename_map()
        self.events.backup_database_start(len(csv_map))
        for table_num, (table, csv_file) in enumerate(csv_map.items(), start=1):
            self.events.backup_table_start(table_num, table)
            self.export_table_to_csv(table, csv_file)
            self.events.backup_table_complete(table_num, table)
        self.events.create_zip_file_start()
        zip_file = self.create_zip_file(csv_map)
        self.remove_csv_files(csv_map)
        self.events.create_zip_file_complete()
        self.events.backup_database_complete()
        return Result.Ok(zip_file)

    def get_table_filename_map(self):
        csv_folder = self.create_csv_folder()
        return {
            DateScrapeStatus: csv_folder.joinpath("scrape_status_date.csv"),
            GameScrapeStatus: csv_folder.joinpath("scrape_status_game.csv"),
            PitchAppScrapeStatus: csv_folder.joinpath("scrape_status_pitch_app.csv"),
            PitchFx: csv_folder.joinpath("pitchfx.csv"),
        }

    def create_csv_folder(self):
        backup_folder = self.config.all_settings.get("DB_BACKUP_FOLDER_PATH")
        backup_folder_path = backup_folder.current_setting(DataSet.ALL).resolve()
        folder_name = f"{datetime.now(timezone.utc).strftime(FILE_TIMESTAMP)}"
        if os.environ.get("ENV") == "TEST":
            folder_name = "__timestamp__"
        csv_folder = Path(backup_folder_path).joinpath(folder_name)
        csv_folder.mkdir(parents=True)
        return csv_folder

    def export_table_to_csv(self, table, csv_file):
        count = last_reported = 0
        total = get_total_number_of_rows(self.db_session, table)
        with csv_file.open("a") as csv:
            for row in table.export_table_as_csv(self.db_session):
                csv.write(f"{row}\n")
                count += 1
                last_reported = self.report_progress(count, total, last_reported)

    def report_progress(self, count, total, last_reported):
        percent_complete = count / float(total)
        if percent_complete - last_reported > ONE_PERCENT:
            self.events.backup_table_progress(percent_complete)
            last_reported = percent_complete
        return last_reported

    def create_zip_file(self, csv_map):
        backup_folder = csv_map.get(DateScrapeStatus).parent
        zip_filename = f"{backup_folder.name}.zip"
        zip_file = backup_folder.parent.joinpath(zip_filename)
        with ZipFile(zip_file, "w", ZIP_DEFLATED) as zip:
            for csv_file in csv_map.values():
                arcname = f"{backup_folder.name}/{csv_file.name}"
                zip.write(csv_file, arcname=arcname)
        return zip_file

    def remove_csv_files(self, csv_map):
        backup_folder = csv_map.get(DateScrapeStatus).parent
        for csv_file in csv_map.values():
            csv_file.unlink()
        backup_folder.rmdir()



"""Shared functions and menus for CLI."""
import subprocess
from functools import partial
from random import randint

import click
from getch import pause

from vigorish.config.database import Season
from vigorish.constants import CLI_COLORS, FIGLET_FONTS
from vigorish.util.result import Result
from vigorish.util.string_helpers import wrap_text


def get_random_cli_color():
    colors = ["red", "blue", "green", "cyan", "magenta", "yellow"]
    return colors[randint(0, len(colors) - 1)]


def get_random_bright_cli_color():
    colors = [
        "bright_red",
        "bright_blue",
        "bright_green",
        "bright_cyan",
        "bright_magenta",
        "bright_yellow",
    ]
    return colors[randint(0, len(colors) - 1)]


def get_random_dots_spinner():
    return f"dots{randint(2, 9)}"


def get_random_figlet_font():
    return FIGLET_FONTS[randint(0, len(FIGLET_FONTS) - 1)]


def print_message(message, wrap=True, max_line_len=70, fg=None, bg=None, bold=None, underline=None):
    if (fg and fg not in CLI_COLORS) or (bg and bg not in CLI_COLORS):
        fg = None
        bg = None
    if wrap:
        message = wrap_text(message, max_len=max_line_len)
    click.secho(message, fg=fg, bg=bg, bold=bold, underline=underline)


def print_heading(message, fg=None, bg=None):
    print_message(f"{message}\n", wrap=False, fg=fg, bg=bg, bold=True, underline=True)


print_error = partial(print_message, fg="bright_red", bg=None, bold=True, underline=None)
print_success = partial(print_message, fg="bright_green", bg=None, bold=True, underline=None)


def validate_scrape_dates(db_session, start_date, end_date):
    result = Season.validate_date_range(db_session, start_date, end_date)
    if result.success:
        season = result.value
        return Result.Ok(season)
    error_heading = "Error! Invalid value for start and/or end dates"
    print()
    print_heading(error_heading, fg="bright_red")
    for s in result.error:
        print_message(s, fg="bright_red")
    print()
    pause(message="Press any key to continue...")
    return Result.Fail("")


def shutdown_cli_immediately():
    subprocess.run(["clear"])
    print_heading("Restart Required!", fg="bright_magenta")
    warning = "Application must be restarted for these changes to take effect."
    print_message(warning, fg="bright_magenta", bold=True)
    print_message("Shutting down vigorish!\n", fg="bright_magenta", bold=True)
    pause(message="Press any key to continue...")
    subprocess.run(["clear"])
    exit(0)



from abc import ABC, abstractmethod


class Task(ABC):
    """Base class for a task that requires access to the db and application state."""

    def __init__(self, app):
        self.app = app
        self.dotenv = app["dotenv"]
        self.config = app["config"]
        self.db_engine = app["db_engine"]
        self.db_session = app["db_session"]
        self.scraped_data = app["scraped_data"]

    @abstractmethod
    def execute(self, *args, **kwargs):
        pass



import subprocess
from collections import Counter
from copy import deepcopy
from datetime import timezone
from pathlib import Path

from events import Events
from halo import Halo

from vigorish.cli.components import get_random_cli_color, get_random_dots_spinner
from vigorish.enums import DataSet, SyncDirection, VigFile
from vigorish.tasks.base import Task
from vigorish.util.datetime_util import dtaware_fromtimestamp
from vigorish.util.regex import URL_ID_REGEX
from vigorish.util.result import Result


class SyncScrapedData(Task):
    def __init__(self, app, cached_s3_objects=None):
        super().__init__(app)
        self._cached_s3_objects = cached_s3_objects
        self.sync_direction = None
        self.file_type = None
        self.data_set = None
        self.year = None
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.events = Events(
            (
                "error_occurred",
                "get_sync_files_start",
                "get_sync_files_complete",
                "sync_files_start",
                "sync_files_progress",
                "sync_files_complete",
            )
        )

    @property
    def file_helper(self):
        return self.scraped_data.file_helper

    @property
    def s3_folderpath_dict(self):
        return self.scraped_data.file_helper.s3_folderpath_dict

    @property
    def cached_s3_objects(self):
        if self._cached_s3_objects:
            return self._cached_s3_objects
        s3_objects = self.file_helper.get_all_object_keys_in_s3_bucket()
        self._cached_s3_objects = [obj for obj in s3_objects]
        return self._cached_s3_objects

    def execute(self, sync_direction, file_type, data_set, year):
        if sync_direction == SyncDirection.UP_TO_S3:
            return self.sync_up_to_s3_bucket(file_type, data_set, year)
        if sync_direction == SyncDirection.DOWN_TO_LOCAL:
            return self.sync_down_to_local_folder(file_type, data_set, year)
        error_message = f"Invalid value for sync_direction: {sync_direction}"
        self.events.error_occurred(error_message)
        return Result.Fail(error_message)

    def sync_up_to_s3_bucket(self, file_type, data_set, year):
        result = self.get_files_to_sync_to_s3(file_type, data_set, year)
        if result.failure:
            self.events.error_occurred("Error occurred analyzing which files need to be synced.")
            return result
        (out_of_sync, new_files, update_files) = result.value
        if not out_of_sync:
            return Result.Ok()
        subprocess.run(["clear"])
        self.upload_files_to_s3(new_files, update_files, file_type, data_set, year)
        return Result.Ok()

    def sync_down_to_local_folder(self, file_type, data_set, year):
        result = self.get_files_to_sync_to_local(file_type, data_set, year)
        if result.failure:
            self.events.error_occurred("Error occurred analyzing which files need to be synced.")
            return result
        (out_of_sync, new_files, update_files) = result.value
        if not out_of_sync:
            return Result.Ok()
        subprocess.run(["clear"])
        self.download_files_to_local_folder(new_files, update_files, file_type, data_set, year)
        return Result.Ok()

    def get_files_to_sync_to_local(self, file_type, data_set, year):
        self.events.get_sync_files_start()
        result = self.get_all_files_stored_locally_and_in_s3(file_type, data_set, year)
        if result.failure:
            return result
        (s3_objects, local_files) = result.value
        if not s3_objects:
            sync_results = (False, [], [])
            self.events.get_sync_files_complete(sync_results)
            return Result.Ok(sync_results)
        sync_results = self.get_files_to_sync(s3_objects, local_files)
        self.events.get_sync_files_complete(sync_results)
        return Result.Ok(sync_results)

    def get_files_to_sync_to_s3(self, file_type, data_set, year):
        self.events.get_sync_files_start()
        result = self.get_all_files_stored_locally_and_in_s3(file_type, data_set, year)
        if result.failure:
            return result
        (s3_objects, local_files) = result.value
        if not local_files:
            sync_results = (False, [], [])
            self.events.get_sync_files_complete(sync_results)
            return Result.Ok(sync_results)
        sync_results = self.get_files_to_sync(local_files, s3_objects)
        self.events.get_sync_files_complete(sync_results)
        return Result.Ok(sync_results)

    def get_all_files_stored_locally_and_in_s3(self, file_type, data_set, year):
        local_files = self.get_all_objects_in_local_folder(file_type, data_set, year)
        s3_objects = self.get_all_objects_in_s3(file_type, data_set, year)
        result = self.check_both_locations_for_duplicates(s3_objects, local_files)
        if result.failure:
            return result
        return Result.Ok((s3_objects, local_files))

    def get_all_objects_in_local_folder(self, file_type, data_set, year):
        folderpath = self.scraped_data.get_local_folderpath(file_type, data_set, year)
        local_files = [
            self.get_local_file_data(file)
            for file in Path(folderpath).glob("*.*")
            if URL_ID_REGEX[file_type][data_set].search(file.stem)
        ]
        return sorted(local_files, key=lambda x: x["name"]) if local_files else []

    def get_all_objects_in_s3(self, file_type, data_set, year):
        s3_objects_dict = {
            VigFile.SCRAPED_HTML: self.get_all_html_objects_in_s3,
            VigFile.PARSED_JSON: self.get_all_json_objects_in_s3,
            VigFile.COMBINED_GAME_DATA: self.get_all_combined_data_objects_in_s3,
            VigFile.PATCH_LIST: self.get_all_patch_list_objects_in_s3,
        }
        return s3_objects_dict[file_type](data_set, year) if file_type in s3_objects_dict else []

    def get_all_json_objects_in_s3(self, data_set, year):
        json_folder = self.s3_folderpath_dict[VigFile.PARSED_JSON][data_set].resolve(year=year)
        html_folder = self.s3_folderpath_dict[VigFile.SCRAPED_HTML][data_set].resolve(year=year)
        s3_objects = [
            self.get_s3_object_data(obj)
            for obj in self.cached_s3_objects
            if json_folder in obj.key
            and html_folder not in obj.key
            and URL_ID_REGEX[VigFile.PARSED_JSON][data_set].search(Path(obj.key).stem)
        ]
        return sorted(s3_objects, key=lambda x: x["name"]) if s3_objects else []

    def get_all_html_objects_in_s3(self, data_set, year):
        html_folder = self.s3_folderpath_dict[VigFile.SCRAPED_HTML][data_set].resolve(year=year)
        s3_objects = [
            self.get_s3_object_data(obj)
            for obj in self.cached_s3_objects
            if html_folder in obj.key
            and URL_ID_REGEX[VigFile.SCRAPED_HTML][data_set].search(Path(obj.key).stem)
        ]
        return sorted(s3_objects, key=lambda x: x["name"]) if s3_objects else []

    def get_all_combined_data_objects_in_s3(self, data_set, year):
        comb_folder = self.s3_folderpath_dict[VigFile.COMBINED_GAME_DATA][DataSet.ALL].resolve(
            year=year
        )
        s3_objects = [
            self.get_s3_object_data(obj)
            for obj in self.cached_s3_objects
            if comb_folder in obj.key
            and URL_ID_REGEX[VigFile.COMBINED_GAME_DATA][DataSet.ALL].search(Path(obj.key).stem)
        ]
        return sorted(s3_objects, key=lambda x: x["name"]) if s3_objects else []

    def get_all_patch_list_objects_in_s3(self, data_set, year):
        json_folder = self.s3_folderpath_dict[VigFile.PARSED_JSON][data_set].resolve(year=year)
        s3_objects = [
            self.get_s3_object_data(obj)
            for obj in self.cached_s3_objects
            if json_folder in obj.key
            and URL_ID_REGEX[VigFile.PATCH_LIST][data_set].search(Path(obj.key).stem)
        ]
        return sorted(s3_objects, key=lambda x: x["name"]) if s3_objects else []

    def get_local_file_data(self, file):
        return {
            "name": file.name,
            "path": file,
            "size": file.stat().st_size,
            "last_modified": dtaware_fromtimestamp(file.stat().st_mtime, use_tz=timezone.utc),
        }

    def get_s3_object_data(self, obj):
        return {
            "name": Path(obj.key).name,
            "path": Path(obj.key),
            "size": obj.size,
            "last_modified": obj.last_modified,
        }

    def check_both_locations_for_duplicates(self, s3_objects, local_files):
        s3_file_names = [f["name"] for f in s3_objects]
        local_file_names = [f["name"] for f in local_files]
        result_s3 = self.ensure_no_duplicates(s3_file_names)
        if result_s3.failure:
            result_s3 = Result.Fail(f"S3 bucket {result_s3.error}")
        result_local = self.ensure_no_duplicates(local_file_names)
        if result_local.failure:
            result_local = Result.Fail(f"Local folder {result_local.error}")
        return Result.Combine([result_s3, result_local])

    def ensure_no_duplicates(self, file_names):
        duplicates = Counter(file_names) - Counter(list(set(file_names)))
        if not duplicates:
            return Result.Ok()
        error = f"contains {len(duplicates)} duplicate file names:\n{', '.join(duplicates)}"
        return Result.Fail(error)

    def get_files_to_sync(self, src_files, dest_files):
        new_files = self.find_files_not_in_dest(src_files, dest_files)
        update_files = self.find_files_newer_in_src(src_files, dest_files)
        if not new_files and not update_files:
            return (False, [], [])
        return (True, new_files, update_files)

    def find_files_not_in_dest(self, src_files, dest_files):
        src_names = set(f["name"] for f in src_files if f["size"] > 0)
        dest_names = set(f["name"] for f in dest_files if f["size"] > 0)
        sync_files = list(src_names - dest_names)
        return [self.add_new_file(f) for f in src_files if f["name"] in sync_files]

    def add_new_file(self, file):
        file["operation"] = "ADD NEW FILE"
        return file

    def find_files_newer_in_src(self, src_files, dest_files):
        (src_sync, dest_sync) = self.find_files_in_common(src_files, dest_files)
        return [
            self.update_existing_file(src_sync[i])
            for i in range(len(src_sync))
            if self.src_file_is_newer(src_sync[i], dest_sync[i])
            and src_sync[i]["size"] != dest_sync[i]["size"]
        ]

    def update_existing_file(self, file):
        file["operation"] = "UPDATE EXISTING FILE"
        return file

    def src_file_is_newer(self, src, dest, min_diff_seconds=60):
        diff = src["last_modified"] - dest["last_modified"]
        return diff.total_seconds() > min_diff_seconds

    def find_files_in_common(self, src_files, dest_files):
        src_names = set(f["name"] for f in src_files if f["size"] > 0)
        dest_names = [f["name"] for f in dest_files if f["size"] > 0]
        sync_files = list(src_names.intersection(dest_names))
        src_sync = [f for f in src_files if f["name"] in sync_files]
        dest_sync = [f for f in dest_files if f["name"] in sync_files]
        src_sync.sort(key=lambda x: x["name"])
        dest_sync.sort(key=lambda x: x["name"])
        return (src_sync, dest_sync)

    def upload_files_to_s3(self, new_files, update_files, file_type, data_set, year):
        sync_files = self.get_sync_files(new_files, update_files)
        self.sync_files(SyncDirection.UP_TO_S3, sync_files, file_type, data_set, year)

    def download_files_to_local_folder(self, new_files, update_files, file_type, data_set, year):
        sync_files = self.get_sync_files(new_files, update_files)
        self.sync_files(SyncDirection.DOWN_TO_LOCAL, sync_files, file_type, data_set, year)

    def get_sync_files(self, new_files, update_files):
        if not new_files and not update_files:
            return []
        sync_files = []
        if new_files:
            sync_files = deepcopy(new_files)
        if update_files:
            sync_files.extend(deepcopy(update_files))
        return sync_files

    def sync_files(self, sync_direction, files, file_type, data_set, year):
        self.bucket_name = self.config.get_current_setting("S3_BUCKET", data_set)
        self.events.sync_files_start(files[0]["name"], 0, len(files))
        for num, file in enumerate(files, start=1):
            self.events.sync_files_progress(file["name"], num - 1, len(files))
            (filepath, s3_key) = self.get_local_and_s3_file_paths(file, file_type, data_set, year)
            self.send_file(sync_direction, filepath, s3_key)
            self.events.sync_files_progress(file["name"], num, len(files))
        self.events.sync_files_complete()

    def get_local_and_s3_file_paths(self, file, file_type, data_set, year):
        local_folder = self.scraped_data.get_local_folderpath(file_type, data_set, year)
        s3_folder = self.s3_folderpath_dict[file_type][data_set].resolve(year=year)
        local_path = Path(local_folder).joinpath(file["name"])
        s3_key = f"{s3_folder}{file['name']}"
        return (str(local_path), s3_key)

    def send_file(self, sync_direction, local_path, s3_key):
        if sync_direction == SyncDirection.UP_TO_S3:
            self.file_helper.client.upload_file(local_path, self.bucket_name, s3_key)
        if sync_direction == SyncDirection.DOWN_TO_LOCAL:
            self.file_helper.resource.Bucket(self.bucket_name).download_file(s3_key, local_path)



from tqdm import tqdm

from vigorish.enums import SeasonType
from vigorish.models.player import Player
from vigorish.models.player_id import PlayerId
from vigorish.models.season import Season
from vigorish.models.team import Team
from vigorish.util.result import Result


def create_relationships(db_session):
    result = link_player_tables(db_session)
    if result.failure:
        return result
    result = link_teams_and_seasons(db_session)
    if result.failure:
        return result

    db_session.commit()
    return Result.Ok()


def link_player_tables(db_session):
    try:
        for p in tqdm(
            db_session.query(Player).all(),
            desc="Linking player tables........",
            unit="row",
            mininterval=0.12,
            maxinterval=5,
            unit_scale=True,
            ncols=90,
        ):
            player_id = PlayerId.find_by_mlb_id(db_session, p.mlb_id)
            if player_id:
                player_id.db_player_id = p.id
        return Result.Ok()
    except Exception as e:
        error = "Error: {error}".format(error=repr(e))
        db_session.rollback()
        return Result.Fail(error)


def link_teams_and_seasons(db_session):
    try:
        for team in tqdm(
            db_session.query(Team).all(),
            desc="Linking team/season tables...",
            unit="row",
            mininterval=0.12,
            maxinterval=5,
            unit_scale=True,
            ncols=90,
        ):
            reg_season = Season.find_by_year(db_session, team.year)
            post_season = Season.find_by_year(
                db_session, team.year, season_type=SeasonType.POST_SEASON
            )
            if reg_season:
                team.regular_season_id = reg_season.id
            if post_season:
                team.post_season_id = post_season.id
        return Result.Ok()
    except Exception as e:
        error = "Error: {error}".format(error=repr(e))
        db_session.rollback()
        return Result.Fail(error)
